#!/usr/bin/tclsh

proc skip_whitespace { s index } {
    set s_trimmed [string trimleft [string range $s $index end] ]
    return [expr [string length $s] - [string length $s_trimmed] ]
}

proc parse_json_object { json index result_var } {
    upvar $result_var result
    set result [list]
    
    if { [string index $json $index] != "\{" } {
        return -1
    }
    incr index
    
    set index [skip_whitespace $json $index]
    
    while { ($index >= 0) && ($index < [string length $json])} {
        set key ""
        set index [parse_json_string $json $index key]
        if { $index < 0 } {
            return -1
        }
        
        set index [skip_whitespace $json $index]
        if { "[string index $json $index]" != ":" } {
            return -1
        }
        incr index
        set value ""
        set index [parse_json_var $json $index value]
        if { $index < 0 } {
            return -1
        }
        lappend result $key
        lappend result $value
        set index [skip_whitespace $json $index]
        if { "[string index $json $index]" == "\}" } {
            incr index
            break
        }
        if { "[string index $json $index]" != "," } {
            return -1
        }
        incr index
        set index [skip_whitespace $json $index]
    }
    return $index
}

proc parse_json_array { json index result_var } {
    upvar $result_var result
    set result [list]
    
    if { [string index $json $index] != "\[" } {
        return -1
    }
    incr index
    
    set index [skip_whitespace $json $index]
    
    while { ($index >= 0) && ($index < [string length $json]) } {
        set value ""
        set index [parse_json_var $json $index value]
        if { $index < 0 } {
            return -1
        }

        lappend result $value
        set index [skip_whitespace $json $index]
        if { "[string index $json $index]" == "\]" } {
            incr index
            break
        }
        if { "[string index $json $index]" != "," } {
            return -1
        }
        incr index
    }
    return $index
}

proc parse_json_string { json index result_var } {
    upvar $result_var result
    set arr ""
    
    if { [string index $json $index] != "\"" } {
        return -1
    }
    incr index
    
    set index [skip_whitespace $json $index]
    
    set end_index [string first "\"" $json $index]
    if { $end_index < 0 } {
        return -1
    }
    
    set result [string range $json $index [expr $end_index - 1] ]
    
    return [expr $end_index + 1]
}

proc parse_json_literal { json index result_var } {
    upvar $result_var result
    set arr ""
    
    set index [skip_whitespace $json $index]
    
    if { [string first "true" $json $index] == $index } {
        set result 1
        return [expr $index + 4]
    }
    
    if { [string first "false" $json $index] == $index } {
        set result 0
        return [expr $index + 5]
    }
    
    if { [string first "null" $json $index] == $index } {
        set result ""
        return [expr $index + 4]
    }
    
    if { [regexp {^-?(?:[0-9]*\.)?[0-9]+(?:[eE][+\-][0-9]+)?} [string range $json $index end] number] } {
        set result $number
        return [expr $index + [string length $number]]
    }
    puts "error parsing literal [string range $json $index [expr $index + 32]]"
    return -1
    set end_index [string first "," $json $index]
    if { $end_index < 0 } {
        return -1
    }
    set result [string range $json $index [expr $end_index - 1] ]
    return $end_index
}


proc parse_json_var { json index result_var } {
    upvar $result_var result
    
    set index [skip_whitespace $json $index]
    switch -- [string index "$json" $index] {
        "\{" {
            set index [parse_json_object "$json" $index result]
        }
        "\[" {
            set index [parse_json_array "$json" $index result]
        }
        "\"" {
            set index [parse_json_string "$json" $index result]
        }
        default {
            set index [parse_json_literal "$json" $index result]
        }
    }
    set index [skip_whitespace "$json" $index]
    return $index
}

proc parse_json { json } {
    set result ""
    parse_json_var "$json" 0 result
    return $result
}

proc escape_json_string { s } {

    global JSON_ESCAPE_MAP
    
    if { ![info exist JSON_ESCAPE_MAP] } {
        set JSON_ESCAPE_MAP [list "\\" "\\\\" "\"" "\\\"" "/" "\\/" "\b" "\\b" "\t" "\\t" "\n" "\\n" "\f" "\\f" "\r" "\\r"]
        lappend JSON_ESCAPE_MAP "\x00" "\\u0000"
        lappend JSON_ESCAPE_MAP "\x01" "\\u0001"
        lappend JSON_ESCAPE_MAP "\x02" "\\u0002"
        lappend JSON_ESCAPE_MAP "\x03" "\\u0003"
        lappend JSON_ESCAPE_MAP "\x04" "\\u0004"
        lappend JSON_ESCAPE_MAP "\x05" "\\u0005"
        lappend JSON_ESCAPE_MAP "\x06" "\\u0006"
        lappend JSON_ESCAPE_MAP "\x07" "\\u0007"
        lappend JSON_ESCAPE_MAP "\x0b" "\\u000b"
        lappend JSON_ESCAPE_MAP "\x0e" "\\u000e"
        lappend JSON_ESCAPE_MAP "\x0f" "\\u000f"
        lappend JSON_ESCAPE_MAP "\x10" "\\u0010"
        lappend JSON_ESCAPE_MAP "\x11" "\\u0011"
        lappend JSON_ESCAPE_MAP "\x12" "\\u0012"
        lappend JSON_ESCAPE_MAP "\x13" "\\u0013"
        lappend JSON_ESCAPE_MAP "\x14" "\\u0014"
        lappend JSON_ESCAPE_MAP "\x15" "\\u0015"
        lappend JSON_ESCAPE_MAP "\x16" "\\u0016"
        lappend JSON_ESCAPE_MAP "\x17" "\\u0017"
        lappend JSON_ESCAPE_MAP "\x18" "\\u0018"
        lappend JSON_ESCAPE_MAP "\x19" "\\u0019"
        lappend JSON_ESCAPE_MAP "\x1a" "\\u001a"
        lappend JSON_ESCAPE_MAP "\x1b" "\\u001b"
        lappend JSON_ESCAPE_MAP "\x1c" "\\u001c"
        lappend JSON_ESCAPE_MAP "\x1d" "\\u001d"
        lappend JSON_ESCAPE_MAP "\x1e" "\\u001e"
        lappend JSON_ESCAPE_MAP "\x1f" "\\u001f"
    }

    return [string map $JSON_ESCAPE_MAP $s]
}

